@@ -14,12 +14,6 @@
 #ifndef FUSE_COMMON_H_
 #define FUSE_COMMON_H_
 
-#ifdef HAVE_LIBFUSE_PRIVATE_CONFIG_H
-#include "fuse_config.h"
-#endif
-
-#include "libfuse_config.h"
-
 #include "fuse_opt.h"
 #include "fuse_log.h"
 #include <stdint.h>
@@ -29,7 +23,7 @@
 #define FUSE_MAJOR_VERSION 3
 
 /** Minor version of FUSE library interface */
-#define FUSE_MINOR_VERSION 14
+#define FUSE_MINOR_VERSION 13
 
 #define FUSE_MAKE_VERSION(maj, min)  ((maj) * 100 + (min))
 #define FUSE_VERSION FUSE_MAKE_VERSION(FUSE_MAJOR_VERSION, FUSE_MINOR_VERSION)
